/* calcul.c - Calcul des positions des joueurs - Detection des collision
 * ----------------------------------------------------------------------------
 * Ce fichier contient toutes les fonctions ayant trait au calcul des positions
 * des joueurs, detection des collisions, detection des tours.
 */

#include <stdio.h>
#include "calcul.h"



/* Structure permettant de stocker
 * l'equation d'une droite
 * (coefficient directeur et constante
 */
struct droite
{
	double coefDir;
	double cte;
};


/* Calcul la position suivante d'un joueur passee en parametre
 * et d'une acceleration. La fonction tient compte des
 * eventuelles collisions avec le decor du circuit.
 *
 * Retourne -> -1 dans le cas normal
 * 	    -> le numero du joueur si il a fait un tour complet
 *          -> -2 en cas d'erreur (acceleration invalide)
 */
int calcul(int coup, struct joueur* tel_j, struct circuit* tel_circuit)
{
	int pos_pred_y, pos_pred_x;		
	int pn2_x, pn2_y, pn2_vx, pn2_vy;
	int acc_lig, acc_col;
	int incx, incy;
	int bool=1, x, y, px, py;
	struct droite d;
	int tour;

	tour=-1;

	/* Verification de l'acceleration */
	if(coup<1 || coup>9)
		return -2;

	/* Calcul de l'acceleration */
	acc_lig=retourne_accel_lig(coup);
	acc_col=retourne_accel_col(coup);

	/* Calcul de la position suivante */
	pn2_y = tel_j->lig_cur + tel_j->v_cur_lig + acc_lig;
	pn2_x = tel_j->col_cur + tel_j->v_cur_col + acc_col;

	/* Calcul de la vitesse suivante */
	pn2_vy = tel_j->v_cur_lig + acc_lig;
	pn2_vx = tel_j->v_cur_col + acc_col;	

	/* Gestion des collisions avec le decor - Detection des tours */
	if(pn2_x==tel_j->col_cur)
	{
		/*
		 * Droite de type x=a.
		 * On verifie qu'aucun point route ne se trouve entre
		 * coordJ->y et pn2.y inclus
		 */
		incy = tel_j->lig_cur <= pn2_y	? 1 : -1;
				
		y=tel_j->lig_cur;
		while(y!=pn2_y+incy)
		{
			if(verifie_tour(y, tel_j->col_cur, tel_circuit, tel_j)>-1)
				tour=tel_j->num;

			if(tel_circuit->trace[y][tel_j->col_cur] == CHAR_DECOR)
			{
				tel_j->lig_cur	= y-incy;
				tel_j->v_cur_col	= 0;
				tel_j->v_cur_lig	= 0;
				bool=0;
				break;
			}
			
			y+=incy;
		}
		
		if(bool)
		{
			/* Deplacement OK */
			tel_j->lig_cur	= pn2_y;
			tel_j->v_cur_col= pn2_vx;
			tel_j->v_cur_lig= pn2_vy;
		}
	}
	else
	{
		d.coefDir = ((double)tel_j->lig_cur -
				 (double)pn2_y) / ((double)tel_j->col_cur - (double)pn2_x);
		d.cte	  = (double)tel_j->lig_cur - d.coefDir*(double)tel_j->col_cur;
		
		incx	= tel_j->col_cur <= pn2_x ? 1 : -1;
		incy	= tel_j->lig_cur <= pn2_y ? 1 : -1;
		
		/*
		 * On va de tel_j->lig_cur a pn2_y INCLUS, d'ou pn2_y+incy,
		 * sinon la boucle s'arrete un point avant la fin.
		 * (Meme remarque pour la seconde boucle while.)
		 */
		y=tel_j->lig_cur;
		while(y!=pn2_y+incy && bool)
		{
			x=tel_j->col_cur;
			while(x!=pn2_x+incx && bool)
			{
				/* Position de x,y par rapport a la droite d */
				px = (double)(((double)y - d.cte)/d.coefDir - (double)x);
				py = (double)(d.coefDir*(double)x +d.cte - (double)y);
				
				px = px<0.0f ? (double)((-1)*px) : px;
				py = py<0.0f ? (double)((-1)*py) : py;
				
				if(px<=0.5f || py<=0.5f)
				{
					if(verifie_tour(y, x, tel_circuit, tel_j)>-1)
						tour=tel_j->num;

					/* Point proche */
					if(tel_circuit->trace[y][x] == CHAR_DECOR)
					{
						tel_j->col_cur=pos_pred_x;
						tel_j->lig_cur=pos_pred_y;
						tel_j->v_cur_col=0;
						tel_j->v_cur_lig=0;
						bool=0;
					}
					
					pos_pred_x = x;
					pos_pred_y = y;
				}
								
				x+=incx;
			}

			y+=incy;
		}
		
		if(bool)
		{
			/* C'est OK! */
			tel_j->lig_cur	= pn2_y;
			tel_j->col_cur	= pn2_x;
			tel_j->v_cur_lig= pn2_vy;
			tel_j->v_cur_col= pn2_vx;
		}
	}

	return tour;
}

/* Retourne l'acceleration en ligne selon l'entier passe en parametre */
int retourne_accel_col(int num)
{
	switch(num)
	{
		case 1: case 4: case 7: return -1;
		case 2: case 5: case 8: return 0;
		case 3: case 6: case 9: return 1;
	}
	return -2;
}

/* Retourne l'acceleration en colonne selon l'entier passe en parametre */
int retourne_accel_lig(int num)
{
	switch(num)
	{
		case 1: case 2: case 3: return 1;
		case 4: case 5: case 6: return 0;
		case 7: case 8: case 9: return -1;
	}
	return -2;
}

/* Retourne -> le numero du checkpoint si s'en est un
 *          -> 0 si c'est la ligne de depart
 *          -> -1
 */
int retourne_checkpoint(int lig, int col, struct circuit* tel_circuit)
{
	int i;

	for(i=0;i<tel_circuit->taille_avant;i++)
	{
		if(lig == tel_circuit->lig_avant[i*2] && col == tel_circuit->lig_avant[i*2+1])
			return tel_circuit->num_lig_avant;
	}
	
	for(i=0;i<tel_circuit->taille_apres;i++)
	{
		if(lig == tel_circuit->lig_apres[i*2] && col == tel_circuit->lig_apres[i*2+1])
			return tel_circuit->num_lig_apres;
	}

	for(i=0;i<tel_circuit->max_joueur;i++)
	{
		if(lig == tel_circuit->pos_dep[i*2] && col == tel_circuit->pos_dep[i*2+1])
			return 0;
	}

	return -1;	
}

/* Empile sans doublons les valeurs passees en parametres
 * Ces valeurs representent les numeros des checkpoint
 * passes par la joueur
 */
void empile_checkpoint(struct joueur* tel_j, int val)
{
	if(tel_j->ind_check==-1)
	{
		tel_j->ind_check=0;
		tel_j->checkpoint[tel_j->ind_check++]=val;
	}
	else
	{
		if(tel_j->checkpoint[0]!=val && tel_j->checkpoint[1]!=val && tel_j->checkpoint[tel_j->ind_check]<2)
		{
			tel_j->checkpoint[tel_j->ind_check++]=val;
		}
	}
}

/* Vide la liste des checkpoints passes par le joueur 
 * (le tableau a toutes ses cases initialisees
 * a la valeur -1
 */
void vide_checkpoint(struct joueur* tel_j)
{
	int i;

	for(i=0;i<2;i++)
		tel_j->checkpoint[i]=-1;

	tel_j->ind_check=-1;

}

/* Retourne -> le numero du joueur qui a fait un tour
 *          -> -1 sinon
 */
int valide_tour(struct joueur* tel_j)
{
	/* Valide tour sans sens */
	/*if((tel_j->checkpoint[0] == 1 && tel_j->checkpoint[1] == 2) || (tel_j->checkpoint[0] == 2 && tel_j->checkpoint[1] == 1))
	{
		tel_j->nb_tours++;

		#ifdef DEBUG
		printf("joueur %d : %d tour(s)\n",tel_j->num,tel_j->nb_tours);
		#endif

		return tel_j->num;
	}*/

	/* Valide tour avec sens */
	if(tel_j->checkpoint[0] == 1 && tel_j->checkpoint[1] == 2)
	{
		tel_j->nb_tours++;

		#ifdef DEBUG
		printf("joueur %d : %d tour(s)\n",tel_j->num,tel_j->nb_tours);
		#endif

		return tel_j->num;
	}

	return -1;
}
	
/* Verifie si le joueur a fait un tour
 * Retourne -> le numero du joueur si celui ci a fait un tour complet
 *          -> -1 sinon
 */
int verifie_tour(int y, int x, struct circuit* tel_circuit, struct joueur* tel_j)
{
	int check;
	int num;

	num=-1;

	if((check=retourne_checkpoint(y,x,tel_circuit))!=-1)
	{
		/* On passe sur la ligne de départ */
		if(check==0)
		{
			num=valide_tour(tel_j);
			vide_checkpoint(tel_j);
		}

		/* On passe sur un checkpoint */
		else if(check>0) empile_checkpoint(tel_j,check);
	}

	return num;
}
