/* maitre.c - Le maitre du jeu
 * ----------------------------------------------------------------------------
 * Quelques conditions :
 *   - Il ne pourra pas y avoir plus de joueurs que de + sur la ligne de depart
 *   - Pour garantir l'atomicite des ecritures dans les tubes, tous le paquets
 *     envoyes seront de meme taille : TAILLE_BUFFER (myfunctions.h)
 *   - Le sens de circulation depend de la position de la ligne de depart
 *   - La ligne de depart doit etre horizontale ou verticale, mais pas en biais
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>

#include "calcul.h"
#include "circuit_struct.h"
#include "joueur_struct.h"
#include "maitre.h"
#include "myfunctions.h"
#include "protocole.h"

char buf[TAILLE_BUFFER+1];


/* Descripteur du tube maitre;  En global, pour la fonction
 * de traitement du signal dont le prototype ne peut être change
 */
int fd_maitre=-1; 


int main(int argc, char** argv)
{
	int nb_joueurs;			/* Le nombre de joueurs    */
	struct joueur* les_joueurs;	/* Liste des joueurs	   */
	struct circuit le_circuit;	/* Le circuit              */
	int nb_tours;
	int i;

	/* Init */
	les_joueurs=NULL;
	
	/******************************
	 * VERIFICATION DES ARGUMENTS *
         ******************************/
	if(argc!=3)
	{
		fprintf(stderr, "usage : %s nb_tours fichier.cir\n", argv[0]);
		return ERROR_ARGS;
	}

	if((nb_tours=atoi(argv[1]))<1)
		sortie_erreur("nb de tours doit etre supérieur ou egal a 1\n",ERROR_TOURS);

	if(access(argv[2], F_OK|R_OK)==-1)
		sortie_erreur("le circuit n'existe pas ou n'est pas accessible\n", ERROR_ACCESS);

	/* Affichage du pid du processus sur la sortie standard */
	printf("%d\n", getpid());

	/* Lecture du circuit dans la structure passee en parametre */
	lecture_circuit(argv[2], &le_circuit);

	/***************************
	 * CREATION DU TUBE MAITRE *
         ***************************/
	sprintf(buf,PATTERN_MAITRE,getpid());
	if(creation_fifo(buf,0772)==-1)
		sortie_erreur("Erreur creation du tube maitre\n", ERROR_CREAT_FIFO);
	else
		if((fd_maitre=ouvrir_fifo(buf,O_RDWR))==-1)
			sortie_erreur("Erreur ouverture du tube maitre\n", ERROR_OPEN_FD);

	/***************************
	 * ACCEPTATION DES JOUEURS *
	 ***************************/
	les_joueurs=accept_joueurs(&le_circuit,&nb_joueurs,TEMPS_CONNECTION);

	/********************
	 * COMPTE A REBOURS *
	 ********************/
	envoie_compte_rebours(les_joueurs, nb_joueurs, TEMPS_REBOURS);

	/*******
	 * JEU *
         *******/
	jeu(les_joueurs, nb_joueurs, &le_circuit, TEMPS_RECEPTION, nb_tours);

	/* Fermeture des tubes et suppression du tube maitre */
	close(fd_maitre);
	sprintf(buf,PATTERN_MAITRE,getpid());
	unlink(buf);

	/* Supression des tubes joueurs et affichage si les fils
 	 * se sont mal termines
	 */
	for(i=0;i<nb_joueurs;i++)
	{
		close(les_joueurs[i].fd_joueur);
		sprintf(buf,PATTERN_JOUEUR,les_joueurs[i].pid_joueur);
		unlink(buf);

		close(les_joueurs[i].fd_aff);
		sprintf(buf,PATTERN_AFFICHE,les_joueurs[i].pid_aff);
		unlink(buf);
	}

	/* Nettoyage des malloc dans les structures */
	for(i=0;i<le_circuit.haut;i++)
		my_free(le_circuit.trace[i]);

	my_free(le_circuit.trace);
	my_free(le_circuit.pos_dep);
	my_free(le_circuit.lig_avant);
	my_free(le_circuit.lig_apres);
	my_free(les_joueurs);

	return NO_ERROR;
}

/* Traitement du signal d'alarme de connection
 * (si aucun joueur connecte apres un temps TEMPS)
 * alors les nouvelles connexions de joueurs sont
 * interrompues
 */
void timeout_connection(int signo)
{
	my_write(fd_maitre,TIMEOUT,TAILLE_PAQUET);
}

/* Traite la requete de connection d'un joueur
 * Retourne : -> la nouvelle structure modifiee si tout se passe bien
 *            -> NULL si le joueur s'est deconnecte pendant le traitement
 */
struct joueur* traite_joueur(struct joueur* tels_j, int nb_j, struct circuit* tel_circuit)
{
	int i;

	/* Allocation - Reallocation */
	tels_j=(struct joueur*)my_realloc(tels_j,sizeof(struct joueur)*(nb_j+1));
	
	/* Ecriture des pids recus */
	sscanf(buf,P_CONNEXION_JOUEUR,&tels_j[nb_j].pid_joueur, &tels_j[nb_j].pid_aff);

	/* Ouverture du tube joueur */
	sprintf(buf,PATTERN_JOUEUR,tels_j[nb_j].pid_joueur);
	if((tels_j[nb_j].fd_joueur=ouvrir_fifo(buf,O_WRONLY))==-1)
		return NULL;

	/* Ouverture du tube affichage */
	sprintf(buf,PATTERN_AFFICHE,tels_j[nb_j].pid_aff);
	if((tels_j[nb_j].fd_aff=ouvrir_fifo(buf,O_WRONLY))==-1)
		return NULL;

	/* Ecriture du numero du joueur dans le tube cv_joueur.pid */
	tels_j[nb_j].num=nb_j;
	sprintf(buf,P_NUM_JOUEUR,tels_j[nb_j].num);
	if(write(tels_j[nb_j].fd_joueur,buf,TAILLE_PAQUET)==-1)
		return NULL;

	/* Attribution d'une position sur la ligne de depart */
	tels_j[nb_j].lig_cur=tel_circuit->pos_dep[nb_j*2];
	tels_j[nb_j].col_cur=tel_circuit->pos_dep[nb_j*2+1];

	/* Envoie du circuit a ce joueur */
	if(envoie_circuit(tel_circuit, tels_j[nb_j].fd_aff)==-1)
		return NULL;

	/* Envoie de sa position a ce joueur par le tube d'affichage */
	sprintf(buf,P_POS_JOUEUR,nb_j,tels_j[nb_j].lig_cur,tels_j[nb_j].col_cur);

	if(write(tels_j[nb_j].fd_aff,buf,TAILLE_PAQUET)==-1)
		return NULL;

	/* Envoie de la position a tous les autres joueurs */
	sprintf(buf,P_POS_AUTRES_JOUEURS,nb_j,tels_j[nb_j].lig_cur, tels_j[nb_j].col_cur);
	for(i=0;i<nb_j;i++)
		my_write_joueur(tels_j,i,tels_j[i].fd_aff,buf,TAILLE_PAQUET);

	/* Envoie des positions precedentes a ce joueur */
	for(i=0;i<nb_j;i++)
	{
		if(tels_j[i].fd_aff!=-1)
		{
			sprintf(buf,P_POS_AUTRES_JOUEURS,i,tels_j[i].lig_cur,tels_j[i].col_cur);
			if(write(tels_j[nb_j].fd_aff,buf,TAILLE_PAQUET)==-1)
				return NULL;
		}
	}

	#ifdef DEGUG
	printf("pid joueur : %d , pid aff : %d, num : %d,\n",tels_j[nb_j].pid_joueur,tels_j[nb_j].pid_aff, nb_j);
	#endif

	return tels_j;
}

/* Envoie le circuit au tube passe en parametre
 * Retourne : -> 0 en cas de succes
 *            -> -1 en cas d'erreur d'ecriture dans le tube
 */
int envoie_circuit(struct circuit* tel_circuit, int tel_fd_aff)
{
	int i,j,k;

	/* Envoie de la taille du circuit de la forme M|LARGEUR|HAUTEUR */
	sprintf(buf,P_TAILLE_CIRCUIT,tel_circuit->larg,tel_circuit->haut);
	if(write(tel_fd_aff, buf, TAILLE_PAQUET)==-1)
		return -1;

	/* Ecriture ligne par ligne du circuit */
	for(i=0;i<tel_circuit->haut;i++)
	{	
		j=0;
		for(k=0;k<tel_circuit->larg;)
		{
			
			for(j=0;j<TAILLE_PAQUET-1 && k<tel_circuit->larg;j++)
				buf[j]=tel_circuit->trace[i][k++];
	
			buf[j]='\0';

			if(write(tel_fd_aff, buf, TAILLE_PAQUET)==-1)
				return -1;
		}
	}

	strcpy(buf,P_FIN_CIRCUIT);
	if(write(tel_fd_aff, buf, TAILLE_PAQUET)==-1)
		return -1;

	return 0;
}

/* Lecture du fichier circuit dans un buffer.
 * La fonction memorise la position de la ligne de depart
 * et calcule les checkpoints virtuels pour la detection
 * des tours.
 * En cas d'erreur de lecture, le programme est quitte
 */
void lecture_circuit(char* nom_circuit, struct circuit* tel_circuit)
{
	char* buf_dyn;
	int i,j;
	int debut;
	int fd_circuit;
	int lig, col, taille;
	int lig_i,col_i;
	struct stat info_fich;

	lig_i=0;
	col_i=0;
	tel_circuit->taille_avant=0;
	tel_circuit->taille_apres=0;
	tel_circuit->pos_dep=NULL;
	tel_circuit->lig_avant=NULL;
	tel_circuit->lig_apres=NULL;

	/* Ouverture du fichier circuit */
	if((fd_circuit=open(nom_circuit,O_RDONLY))==-1)
		sortie_erreur("Erreur ouverture circuit\n",ERROR_OPEN_FD);

	/* Recupere la taille du fichier et allocation du buffer correspondant */
	if(fstat(fd_circuit,&info_fich)==-1)
		sortie_erreur("Erreur stat\n",ERROR_STAT);

	buf_dyn=(char*)my_malloc((info_fich.st_size)*sizeof(char));
	
	/* Lecture du circuit dans un buffer */
	my_read(fd_circuit,buf_dyn,info_fich.st_size);
	close(fd_circuit);

	/* Analyse et stockage des dimensions du circuit dans la structure */
	sscanf(buf_dyn,"%d%*[ \t\n\r\v\f]%d",&tel_circuit->larg,&tel_circuit->haut);

	/* Allocation du tableau a double dimension pour le circuit */
	tel_circuit->trace=(char**)my_malloc(tel_circuit->larg*tel_circuit->haut*sizeof(char));
	for(i=0;i<tel_circuit->haut;i++)
		tel_circuit->trace[i]=(char*)my_malloc(tel_circuit->larg*sizeof(char));

	/* On se positionne au 1er caractere du circuit */
	debut=0;
	while(buf_dyn[debut]!=CHAR_ROUTE && buf_dyn[debut]!=CHAR_DEP_ARR && buf_dyn[debut]!=CHAR_DECOR)
		debut++;

	/* On ecrit le circuit dans la structure */
	i=debut; j=0; tel_circuit->max_joueur=0; tel_circuit->pos_dep=NULL;
	taille=tel_circuit->larg*tel_circuit->haut;
	while(j!=taille)
	{
		lig=(j-j%tel_circuit->larg)/tel_circuit->larg;
		col=j%tel_circuit->larg;

		/* Memorise les differentes positions de la ligne de depart */
		if(buf_dyn[i]==CHAR_DEP_ARR)
		{	
			tel_circuit->pos_dep=(int*)my_realloc(tel_circuit->pos_dep,(tel_circuit->max_joueur+1)*sizeof(int)*2);
			tel_circuit->pos_dep[tel_circuit->max_joueur*2]=lig;
			tel_circuit->pos_dep[tel_circuit->max_joueur*2+1]=col;

			/* Pour determiner le sens (horizontal ou vertical) de la ligne de depart */
			if(tel_circuit->max_joueur>=1)
			{
				if(lig==tel_circuit->pos_dep[(tel_circuit->max_joueur-1)*2])  lig_i++;
				if(col==tel_circuit->pos_dep[(tel_circuit->max_joueur-1)*2+1])col_i++;
			}

			tel_circuit->max_joueur++;

		}

		/* Ignore les retours a la ligne - Copie le circuit dans le
		 * tableau a double dimensions
		 */
		if(buf_dyn[i]!='\n' || buf_dyn[i]==CHAR_DECOR || buf_dyn[i]==CHAR_ROUTE || buf_dyn[i]==CHAR_DEP_ARR)
		{	
			tel_circuit->trace[lig][col]=buf_dyn[i];
			j++;
		}

		i++;
	}

	/* Calcul des positions des 2 checkpoint virtuels */
	for(i=0;i<tel_circuit->max_joueur;i++)
	{	
		/* Ligne horizontale */
		if(lig_i==tel_circuit->max_joueur-1)
		{
			lig=tel_circuit->pos_dep[i*2]-1;
			col=tel_circuit->pos_dep[i*2+1];
		}

		/* Ligne verticale */
		if(col_i==tel_circuit->max_joueur-1)
		{
			lig=tel_circuit->pos_dep[i*2];
			col=tel_circuit->pos_dep[i*2+1]+1;
		}	

		/* Apres la ligne */
	
		tel_circuit->num_lig_apres=1;
		if(tel_circuit->trace[lig][col]==CHAR_ROUTE )
		{
			tel_circuit->lig_apres=(int*)my_realloc(tel_circuit->lig_apres,(tel_circuit->taille_apres+1)*2*sizeof(int));
			tel_circuit->lig_apres[tel_circuit->taille_apres*2]=lig;
			tel_circuit->lig_apres[tel_circuit->taille_apres*2+1]=col;		
			tel_circuit->taille_apres++;
		}

		/* Avant la ligne */

		/* Ligne horizontale */
		if(lig_i==tel_circuit->max_joueur-1)
		{
			lig=tel_circuit->pos_dep[i*2]+1;
		}

		/* Ligne verticale */
		if(col_i==tel_circuit->max_joueur-1)
		{
			col=tel_circuit->pos_dep[i*2+1]-1;
		}

		tel_circuit->num_lig_avant=2;
		if(tel_circuit->trace[lig][col]==CHAR_ROUTE)
		{
			tel_circuit->lig_avant=(int*)my_realloc(tel_circuit->lig_avant,(tel_circuit->taille_avant+1)*2*sizeof(int));
			tel_circuit->lig_avant[tel_circuit->taille_avant*2]=lig;
			tel_circuit->lig_avant[tel_circuit->taille_avant*2+1]=col;		
			
			tel_circuit->taille_avant++;
		}		
}

	/* Desallocation */
	my_free((void*)buf_dyn);
}

/* Supprime un joueur du jeu
 * Le joueur n'est pas reellement supprime, mais les champs le
 * representant se voient affectes des valeurs qui feront que
 * ce joueur sera ignore par la suite
 */
void supprime_joueur_jeu(struct joueur* tels_j, int num_j)
{
	close(tels_j[num_j].fd_joueur);
	close(tels_j[num_j].fd_aff);

	/*
	tels_j[num_j].pid_joueur=0;
	tels_j[num_j].pid_aff	=0;*/
	tels_j[num_j].num	=-1;
	tels_j[num_j].fd_joueur	=-1;
	tels_j[num_j].fd_aff	=-1;

	#ifdef DEBUG
	printf("suppr joueur %d\n",num_j);
	#endif
}

/* Ecris au joueur dont le descripteur de tube (fd) est passe en parametre
 * En cas d'erreur d'ecriture, le joueur est supprime du jeu
 * Retourne : -> le nombre de caracteres ecris en cas de succes
 *            -> -1 en cas d'erreur d'ecriture ou d'ecriture ignoree
 */
ssize_t my_write_joueur(struct joueur* tels_j, int num_j, int fd, const void *buf, size_t count)
{
	ssize_t ret;

	if(fd!=-1)
	{
		if((ret=write(fd,buf,count))==-1)
			supprime_joueur_jeu(tels_j,num_j);
	}
	else
		ret=-1;

	return ret;
}

/* Compte le nombre de joueurs reelement en jeu
 *  Retourne -> ce nombre
 */
int retourne_nb_joueur(struct joueur* tels_j, int nb_j)
{
	int i;
	int nb;

	nb=0;
	for(i=0;i<nb_j;i++)
	{
		if(tels_j[i].num!=-1)
			nb++;
	}

	return nb;
}

/* Envoie le compte a rebours a tous les joueurs */
void envoie_compte_rebours(struct joueur* tels_j,int nb_j, int temps)
{
	int i,j;

	/********************
	 * COMPTE A REBOURS *
	 ********************/
	for(i=temps;i>=0;i--)
	{	
		sprintf(buf,P_COMPTE_REBOURS,i);
		for(j=0;j<nb_j;j++)
			my_write_joueur(tels_j,j,tels_j[j].fd_joueur,buf,TAILLE_PAQUET);
		sleep(1);	
	}

	for(i=0;i<nb_j;i++)
		my_write_joueur(tels_j,i,tels_j[i].fd_joueur,P_FIN_COMPTE_REBOURS,TAILLE_PAQUET);
}

/* Attend la connection des joueurs et lance le traitement de ce joueur
 * Retourne -> le tableau de structures de joueurs en cas de succes
 *          -> NULL si il n'y a aucun joueur
 */
struct joueur* accept_joueurs(struct circuit* tel_circuit,int* nb_total_joueur, int temps_attente)
{
	struct joueur* les_j;
	struct joueur* le_j_temp;
	int nb_j;
	int i;

	nb_j=0;
	les_j=NULL;

	/* Time out de connection : si aucun joueur n'est connecte
	 * apres TEMPS_CONNECTION, alors le programme se termine
	 */
	signal(SIGPIPE,SIG_IGN);
	signal(SIGALRM,timeout_connection);
	alarm(temps_attente);

	#ifdef DEBUG
	printf("\nAttente de connexion des joueurs ...\n");
	#endif

	while(read(fd_maitre,buf,TAILLE_PAQUET)>0 && strcmp(buf,TIMEOUT)!=0)
	{
		/* Traitement du joueur qui se connecte */
		le_j_temp=traite_joueur(les_j,nb_j,tel_circuit);
		if(le_j_temp!=NULL) /* NULL : Le joueur s'est deconnecte pdt le traitement */
		{
			les_j=le_j_temp; le_j_temp=NULL;

			/* Incrementation du nombre de joueurs */
			nb_j++;
		}
		else
		{
			les_j=(struct joueur*)my_realloc(les_j,sizeof(struct joueur)*nb_j);
		}

		/* Teste si on atteint le nombre maxi de joueurs */
		if(nb_j==tel_circuit->max_joueur)
		{
			#ifdef DEBUG
			printf("Nombre maximum de joueurs atteint ...\n");
			#endif

			break;
		}
	}

	*nb_total_joueur=nb_j;

	#ifdef DEBUG
	nb_j==0 ? printf("Aucun joueur\n") : printf("%d joueur(s)\n",nb_j);
	#endif

	/* Desactivation de l'alarme */
	signal(SIGALRM,SIG_IGN);

	/* Envoie le signal de fin de reception de coordonnees a tous les joueurs */
	for(i=0;i<nb_j;i++)
		my_write_joueur(les_j,i,les_j[i].fd_aff,P_FIN_POS_AUTRES_JOUEURS,TAILLE_PAQUET);

	return nb_j>0 ? les_j : NULL ;
}

/* Jeu -> les positions des joueurs
 *     -> calcul la nouvelle position de ce joueur
 *     -> envoie cette  position a tous les joueurs
 */
void jeu(struct joueur* les_j, int nb_j, struct circuit* tel_circuit, int temps_idle, int nb_tours)
{
	int i,j,ret;

	int joueur_lu;	/* Le numero de joueur lu  */
	int accel_lu;	/* L'acceleration envoyee  */
	int arrive;

	arrive=0;

	/* Initialise les positions de departs des joueurs */
	for(i=0;i<nb_j;i++)
	{
		les_j[i].v_cur_lig=0;
		les_j[i].v_cur_col=0;
		les_j[i].ind_check=-1;
		les_j[i].nb_tours=0;
		les_j[i].arrive=0;
		for(j=0;j<2;j++)
			les_j[i].checkpoint[j]=-1;
	}

	/* Si on recois rien pendant n secondes, on arrete le jeu */
	signal(SIGALRM,timeout_connection);
	alarm(temps_idle);

	/* Envoie nombre de tours a chaque joueur */
	for(j=0;j<nb_j;j++)
	{
		sprintf(buf,P_TOUR,joueur_lu,1,nb_tours);
		my_write_joueur(les_j,j,les_j[j].fd_aff,buf,TAILLE_PAQUET);
	}

	/* Attente de reception des deplacements des joueurs */
	while(retourne_nb_joueur(les_j, nb_j)>0 && my_read(fd_maitre,buf, TAILLE_PAQUET)>0
						&& strcmp(buf,TIMEOUT)!=0)
	{
		sscanf(buf,P_COUP_JOUEUR,&joueur_lu,&accel_lu);

		/* Si le joueur est deja arrive, on ignore ses deplacements */
		if(est_arrive(&les_j[joueur_lu])==1) continue;

		/* Verification des tours, de l'arrivee */
		ret=calcul(accel_lu, &les_j[joueur_lu], tel_circuit);
		test_tour(ret, &arrive, les_j, nb_j, joueur_lu, nb_tours);

		/* Envoie du deplacement a tous joueurs */
		sprintf(buf,P_NOUV_POS_JOUEUR,joueur_lu,les_j[joueur_lu].lig_cur,les_j[joueur_lu].col_cur);
		for(j=0;j<nb_j;j++)
			my_write_joueur(les_j,j,les_j[j].fd_aff,buf,TAILLE_PAQUET);

		/* Verifie si tout le monde est arrive */
		if(arrive==retourne_nb_joueur(les_j, nb_j))
		{
			#ifdef DEBUG
			printf("Tout le monde arrive --> Fin course\n");
			#endif

			break;

		}

		/* On repositionne l'alarme (les appels a alarme n'etant pas empilables) */
		alarm(temps_idle);		
	}

	/* Ignore l'alarme */
	signal(SIGALRM,SIG_IGN);

	/* Envoie le signal de fin de jeu */
	for(j=0;j<nb_j;j++)
		my_write_joueur(les_j,j,les_j[j].fd_aff, P_FIN_POS_JOUEUR,TAILLE_PAQUET);

	/*************
	 * NETTOYAGE *
         *************/
	for(i=0;i<nb_j;i++)
	{
		close(les_j[i].fd_joueur);
		close(les_j[i].fd_aff);
	}
}

/* Retourne -> 1 si le joueur est arrive
 * 	    -> 0 sinon
 */
int est_arrive(struct joueur* tel_j)
{
	if(tel_j->arrive==1)
		return 1;

	return 0;
}

void test_tour(int ret, int* arrive, struct joueur* les_j, int nb_j, int num_j, int nb_tours)
{
	if(ret>-2)
	{
		/* Cas d'un joueur qui a fait un tour */
		if(ret>-1)
		{
			sprintf(buf,P_TOUR,num_j,les_j[num_j].nb_tours+1,nb_tours);
			my_write_joueur(les_j,num_j,les_j[num_j].fd_aff,buf,TAILLE_PAQUET);

			/* Teste si il gagne */
			if(les_j[num_j].nb_tours==nb_tours)
			{
				les_j[num_j].arrive=1;
					(*arrive)++;
					
				/* Le joueur est 1er */
				if(*arrive==1)
				{
					my_write_joueur(les_j,num_j,
							les_j[num_j].fd_aff,
							P_GAGNE,TAILLE_PAQUET);	
	
					#ifdef DEBUG
					printf("Joueur %d gagne 1er !\n", num_j);
					#endif
				}

				/* Les autres joueurs arrive apres le 1er */
				else
				{
					sprintf(buf,P_ARR,*arrive,nb_j);
					my_write_joueur(les_j,num_j,
							les_j[num_j].fd_aff,
							buf,TAILLE_PAQUET);	

					#ifdef DEBUG
					printf("Joueur %d gagne pos %d\n", num_j, *arrive);
					#endif

				}
			}
		}
	}
}
