/* my_ncurses.c - Fonctions ncurses
 * ----------------------------------------------------------------------------
 * Ce fichier contient plusieurs fonctions utile pour l'affichage ncurses :
 *  - Affichage d'un entier (en gras ou non) a la position voulue
 *  - Affichage d'un caractere, d'une chaine a la position voulue
 *  - Rafraichir un nombre de fenetre quelconques passees en parametres
 *  - Creer une fenetre
 *  - Terminer l'affichage en mode ncurses
 */

#include <string.h>
#include <stdarg.h>
#include "myfunctions.h"
#include "my_ncurses.h"

#include <term.h>

#define ERROR_NC "pb ncurses\n"

char buf[TAILLE_BUFFER+1];

/* Affiche un entier a la position voulue */
void ncurses_print(WINDOW* win, int telEntier, int l, int c)
{
	if(wmove(win,l,c)==ERR)
		sortie_erreur(ERROR_NC,ERROR_NCURSES);

	if(wprintw(win,"%d",telEntier)==ERR)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
}

/* Affiche un entier en gras a la position voulue */
void ncurses_print_bold(WINDOW* win, int telEntier, int l, int c)
{
	if(wattron(win,A_BOLD)==ERR)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);

	ncurses_print(win, telEntier, l, c);

	if(wattroff(win,A_BOLD)==ERR)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
}

/* Affiche un caractere a la position voulue */
void ncurses_print_char(WINDOW* win, char telCar, int l, int c)
{
	if(wmove(win,l,c)==ERR)
		sortie_erreur(ERROR_NC,ERROR_NCURSES);

	if(wprintw(win,"%c",telCar)==ERR)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
}

/* Affiche une chaine de caractere a la position voulue */
void ncurses_print_string(WINDOW* win, char* telleChaine, int l, int c)
{
	if(wmove(win,l,c)==ERR)
		sortie_erreur(ERROR_NC,ERROR_NCURSES);

	if(wprintw(win,"%s",telleChaine)==ERR)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);

}

/* Rafraichit les fenetres passee en parametre */
void ncurses_refresh(int nb, ...)
{
	va_list ap;
	
	va_start(ap, nb);
	
	while(nb--!=0)
	{
		if(wnoutrefresh(va_arg(ap, WINDOW*))==ERR)
			sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);	
	}
	
	va_end(ap);
	
	if(doupdate()==ERR)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
}

/* Rafraichit les fenetres passee en parametre */
void ncurses_clear(int nb, ...)
{
	va_list ap;
	
	va_start(ap, nb);
	
	while(nb--!=0)
	{
		if(wclear(va_arg(ap, WINDOW*))==ERR)
			sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);	
	}
	
	va_end(ap);
	
	if(doupdate()==ERR)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
}

/* Cree une fenetre et la retourne */
WINDOW* ncurses_wcreate(int nlines, int ncols, int begin_y, int begin_x)
{	
	WINDOW* win_circuit;
		
	if((win_circuit=newwin(nlines, ncols, begin_y, begin_x))==NULL)
		sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);

	return win_circuit;		
}

/* Initialise l'affichage ncurses.
 * -> Verifie la taille du terminal
 * -> Enleve le curseur si le terminal en a la capacite
 */
void ncurses_init(int l, int c)
{
	/* Verification de la taille du terminal */
	if(setupterm(NULL,fileno(stdout),NULL)==OK)
	{
		if(tigetnum("lines")<l || tigetnum("cols")<c)
		{
			sprintf(buf,"terminal trop petit : requis : %d x %d\n",l,c);
			sortie_erreur_fils(buf,ERROR_TERM);
		}
	}

	/* Enleve le curseur si possible */
	if(setupterm(NULL,fileno(stdout),NULL)==OK)
	{
		strcpy(buf, tigetstr("civis"));
		if(buf>(char*)0)
			printf(tparm(buf));
	}

	/* Initialisation ncurses */
	if(initscr()==NULL)	sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
	if(noecho()==ERR)	sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
	if(cbreak()==ERR)	sortie_erreur_fils(ERROR_NC,ERROR_NCURSES);
}

/* Termine l'affichage ncurses.
 * -> Remet le curseur
 * -> Supprime les fenetres
 */
void ncurses_end(int nb, ...)
{
	va_list ap;
	va_start(ap, nb);

	/* Remet le curseur si possible */
	if(setupterm(NULL,fileno(stdout),NULL)==OK)
	{
		strcpy(buf, tigetstr("cnorm"));
		if(buf>(char*)0)
			printf(tparm(buf));
	}


	
	while(nb--!=0)
		delwin(va_arg(ap, WINDOW*));
	
	va_end(ap);

	echo(); nocbreak(); endwin();
}
