/* myfunctions.c - Quelques fonction utiles un peu partout
 * ----------------------------------------------------------------------------
 * Dans ce fichier, quelques fonctions sont definies pour simplifier l'ecriture
 * des sources : 
 *  - Une fonction de sortie sur erreur en affichant un message
 *  - Des fonctions de lecture/ecriture avec sortie sur erreur
 *  - Des fonctions pour creer, ouvrir ou tester l'existence de tubes nommes
 *  - Des fonction malloc, realloc et free avec sortie sur erreur
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "myfunctions.h"

/* Sortie sur erreur 
 * Affiche un message d'erreur et quitte le programme en
 * retournant un code d'erreur a l'appelant
 */
void sortie_erreur(char* chaine, int error_code)
{
	fprintf(stderr,chaine);
	exit(error_code);
}

/* Write avec sortie sur erreur
 * Retourne -> le nombre de d'octets ecrits
 * Sort	    -> en cas d'erreur d'ecriture
 */
ssize_t my_write(int fd, const void *buf, size_t count)
{
	ssize_t ret;

	if((ret=write(fd,buf,count))==-1)
		sortie_erreur("Erreur ecriture\n",ERROR_WRITE_FD);

	return ret;
}

/* Read avec sortie sur erreur
 * Retourne -> le nombre de d'octets lus
 * Sort	    -> en cas d'erreur de lecture
 */
ssize_t my_read(int fd, void *buf, size_t count)
{
	ssize_t ret;

	if((ret=read(fd,buf,count))==-1)
		sortie_erreur("Erreur lecture\n",ERROR_READ_FD);

	return ret;
}


/* Cree un tube nomme
 * Retourne -> 0 si le tube est bien cree ou si il existe deja
 *          -> -1 en cas d'erreur de creation ou si un fichier du
 *             meme nom existe (qui n'est donc pas un tube)
 */
int creation_fifo(char* nom_tube, mode_t mode_creation)
{
	if(access(nom_tube,F_OK)==-1)
		return mkfifo(nom_tube,mode_creation);
	else
		return existe_fifo(nom_tube);	
}

/* Ouvrir un tube nomme
 * Retourne -> le descripteur (>=0) en cas de succes
 *          -> -1 si le tube n'existe pas
 */
int ouvrir_fifo(char* nom_tube, int mode_ouverture)
{
	if(existe_fifo(nom_tube)==-1)
		return -1;
	else
		return open(nom_tube,mode_ouverture);
}

/* Teste si un tube nomme existe
 * Retourne -> 0 si il existe
 *          -> -1 si il n'existe pas ou que le fichier n'est pas un tube
 */
int existe_fifo(char* nom_tube)
{
	struct stat info_tube;

	if(stat(nom_tube,&info_tube)==-1)
		return -1;

	if(!S_ISFIFO(info_tube.st_mode))
		return -1;

	return 0;
}

/* Redefinition de malloc
 * Retourne -> le pointeur alloue en cas de succes
 * Sort     -> en cas d'echec
 */
void* my_malloc(size_t size)
{
	void* ptr;

	ptr=(void*)malloc(size);
	
	if(ptr==NULL)
		sortie_erreur("Erreur malloc\n",ERROR_MALLOC);

	return ptr;
}

/* Redefinition de realloc
 * Retourne -> le pointeur alloue en cas de succes
 * Sort     -> en cas d'echec
 */
void* my_realloc(void* ptr, size_t size)
{
	ptr=(void*)realloc(ptr,size);
	
	if(ptr==NULL)
		sortie_erreur("Erreur realloc\n",ERROR_MALLOC);

	return ptr;
}

void my_free(void* ptr)
{
	if(ptr!=NULL)
		free((void*)ptr);
}

